//*****************************************************************************
//
// Random Number test harness
//  Copyright (C) 2010 C Turvey, All rights reserved.
//   clive@tbcnet.com, cturvey@gmail.com
//
//*****************************************************************************
//
// Compiled with MSVC (12.00) Visual Studio 98 :  cl -Ox rndtest.c
//
//*****************************************************************************

#include <windows.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

//#define DEBUG

#define ELEMENTS (80000)
#define RANGE (50)  // Up to 64 should work

#define MATCHBINS (32) // Arbitrarily enough I think

//*****************************************************************************

DWORD *Buffer; // Area to store multiple four number groups, 4 * ELEMENTS

__int64 *BitVector; // Area to store vectorization of number groups, 1 * ELEMENTS

int visited[RANGE]; // Scoreboard for selected numbers, 0 .. (RANGE-1)

int hits[RANGE]; // Distribution table of random numbers, 0 .. (RANGE-1)

DWORD Bins[MATCHBINS]; // Distribution of repeative groups

unsigned int rnd;

//*****************************************************************************

int getrand(void)
{
#if 1 // Maximal Length Sequence 32-bit (2**32 - 1)

  // 6-bits at a time
  rnd = (rnd << 6) | ((rnd ^ (rnd << 2) ^ ((rnd ^(rnd << 1)) << 6)) >> (32 - 6));

  return(rnd & 0x3F);
#else // Library
  return(rand() % RANGE);
#endif
}

//*****************************************************************************

void random4(DWORD *p)
{
  int a, b, c, d;

  // Select four unique number in the range 0..(RANGE-1)

  // visited[] assumed clear on entry

  a = getrand();
  while((a >= RANGE) || visited[a])
    a = getrand();
  visited[a] = 1;

  b = getrand();
  while((b >= RANGE) || visited[b])
    b = getrand();
  visited[b] = 1;

  c = getrand();
  while((c >= RANGE) || visited[c])
    c = getrand();
  visited[c] = 1;

  d = getrand();
  while((d >= RANGE) || visited[d])
    d = getrand();
  visited[d] = 1; // overkill

  visited[a] = 0; // Clear scoreboard for current group
  visited[b] = 0;
  visited[c] = 0;
  visited[d] = 0; // overkill

  hits[a]++; // Account for numbers selected in distribution
  hits[b]++;
  hits[c]++;
  hits[d]++;

  a++;  // Convert from 0..(RANGE-1) to 1..RANGE
  b++;
  c++;
  d++;

  p[0] = a; // Store local selection into global table passed in
  p[1] = b;
  p[2] = c;
  p[3] = d;

#ifdef DEBUG // Output for debugging
  printf("[%2d, %2d, %2d, %2d]\n",a,b,c,d);
#endif
}

//*****************************************************************************

// Read Pentium class time stamp counter (CPU/Instruction cycle counter)
//  The register is 64-bit (edx:eax), we use 32-bits (eax) for convenience.

// For practical purposes if the 32-bit overflows in Windows you are
//  timing too long, and counting cycles billed to other threads/processes

DWORD ReadTSC(void)
{
  DWORD x;

  __asm
  {
    rdtsc
    mov [x],eax
  }

  return(x);
}

//*****************************************************************************

// Quick Sort comparison routine for 64 bit elements, called via qsort()

int cmp64(const void *x, const void *y)
{
  __int64 i = *(__int64 *)x - *(__int64 *)y;

  if (i == 0) // x equal to y
    return(0);
  else if (i > 0)
    return(1); // x bigger than y
  else
    return(-1); // x smaller than y
}

//*****************************************************************************

// Decode bit vector into ASCII for print out, should permit up to 64 bits
//  to be set, in this application only expecting four. Numbers output in
//  ascending order.
// Output in the form [1,2,3,4], or [47,48,49,50]

char *DecodeVec(__int64 x)
{
  static char s[256], *t;
  int i, j;

  j = 0; // flag initial

  t = s; // Initialize string pointer

  i = 1; // Bit index (1..64)

  while(x) // Until data has been fully expressed
  {
    if (x & 1) // is the current bit set?
    {
      if (j++ == 0)
        *t++ = '['; // First open with a square bracket
      else
        *t++ = ','; // Subsequently use a comma

      if (i > 9) // For numbers 10 or more display high order digit (works to 99)
        *t++ = '0' + (i / 10);

      *t++ = '0' + (i % 10); // Display low order digit
    }

    i++; // Next bit index

    x >>= 1; // Move to next bit
  }

  *t++ = ']'; // Close with a square bracket

  *t++ = 0; // Terminate C string with NUL

  return(s);
}

//*****************************************************************************

int main(int argc, char **argv)
{
  DWORD TotalTickStart, TotalTickEnd; // Time stamps for application
  DWORD TotalTSCStart, TotalTSCEnd;

  DWORD TickStart, TickEnd; // Time stamps for sections
  DWORD TSCStart, TSCEnd;

  DWORD *p; // Pointer for DWORD values
  __int64 *q; // Pointer for 64-bit values

  __int64 x;

  int i, j, k, l;

  // Sync to beginning of time quanta

  Sleep(0);

  // Get start time

  TotalTickStart = GetTickCount();
  TotalTSCStart = ReadTSC();

  // Initialize buffers, and random seeds

  rnd = (unsigned)time(NULL);

  srand((unsigned)time(NULL));

  memset(visited, 0, sizeof(visited)); // Clear visited statistics

  memset(hits, 0, sizeof(hits)); // Clear distribution statistics

  // Allocate buffer for multiple 4 number groups, using 32-bit DWORDs

  j = ELEMENTS * 4 * sizeof(DWORD);

  printf("Buffer %10d bytes\n\n", j);

  Buffer = malloc(j);

  // Build the random number table

  printf("Generating..\n\n");

  TickStart = GetTickCount();
  TSCStart = ReadTSC();

  p = Buffer;

  for(i=0; i<ELEMENTS; i++)
  {
    random4(p);

    p += 4;
  }

  TSCEnd = ReadTSC();
  TickEnd = GetTickCount();

  printf("%10d Ticks, %10d Cycles\n",TickEnd - TickStart, TSCEnd - TSCStart);

  putchar('\n');


  // Save the random numbers to a file

  printf("Saving.. 'rndtest.out'\n\n");

  if (1)
  {
    HANDLE hFile;

    hFile = CreateFile(
      "rndtest.out",
	    GENERIC_WRITE,	          // access (read-write) mode
	    0,							          // share mode
	    NULL,						          // pointer to security descriptor
	    CREATE_ALWAYS,				    // how to create
      FILE_ATTRIBUTE_NORMAL,		// file attributes
		  NULL);                    // handle to file with attributes to copy

    if (hFile && (hFile != INVALID_HANDLE_VALUE))
    {
      DWORD dwWrite;

      dwWrite = ELEMENTS * 4 * sizeof(DWORD);

      WriteFile(hFile, (BYTE *)Buffer, dwWrite, &dwWrite, NULL);

      CloseHandle(hFile);
    }
  }

  putchar('\n');


#ifdef DEBUG
  // Output the distribution of random numbers across the desired range

  printf("Evaluating Distribution..\n\n");

  j = (i * 4) / RANGE;

  printf("%8d Expected Average per bin\n\n",j);

  for(i=0; i<RANGE; i++)
  {
    printf("%2d : %10d %6d\n",i+1,hits[i],hits[i]-j);
  }

  putchar('\n');
#endif


  // Allocate buffer for bit vector, using 64-bit variable

  j = ELEMENTS * sizeof(__int64); // Allocate enough memory for the vectorization

  BitVector = malloc(j);


  // Convert the numeric representation into bit fields
  //  this could be folded into the generation, and score boarding
  //  previously selected random numbers.

  printf("Converting to bit vectors..\n\n");

  TickStart = GetTickCount();
  TSCStart = ReadTSC();

  p = Buffer;
  q = BitVector;

  for(i=0; i<ELEMENTS; i++)
  {
    *q = ((__int64)1 << (p[0]-1)) | ((__int64)1 << (p[1]-1)) | ((__int64)1 << (p[2]-1)) | ((__int64)1 << (p[3]-1));

    p += 4;
    q++;
  }

  TSCEnd = ReadTSC();
  TickEnd = GetTickCount();

  printf("%10d Ticks, %10d Cycles\n",TickEnd - TickStart, TSCEnd - TSCStart);

  putchar('\n');


  // Perform a Quick Sort of the vector table using standard C library function
  //  for simplicity of implementation

  printf("Sorting bit vectors..\n\n");

  TickStart = GetTickCount();
  TSCStart = ReadTSC();

  qsort(BitVector, ELEMENTS, sizeof(__int64), cmp64); // sort of 64-bit values

  TSCEnd = ReadTSC();
  TickEnd = GetTickCount();

  printf("%10d Ticks, %10d Cycles\n",TickEnd - TickStart, TSCEnd - TSCStart);

  putchar('\n');


  // Process the vector table counting consecutive matches

  printf("Scanning..\n\n");

  TickStart = GetTickCount();
  TSCStart = ReadTSC();

  memset(Bins, 0, sizeof(Bins)); // Clear Bins

#ifdef DEBUG
  for(i=0; i<10; i++) // Show first 10 sorted vectors for visual check
    printf("%016I64X\n",BitVector[i]);
#endif

  q = BitVector; // Pointer into table

  x = *BitVector; // Current value

  j = 0; // Count at current, will get incremented as entry 1 will match entry 1

  k = 0; // Count of unique

  l = 0; // Count of total (crosschecking)

  for(i=0; i<ELEMENTS; i++)
  {
    if (x == *BitVector)
    {
      j++; // Increment Matches
    }
    else // Bit vector changed
    {
      k++; // Increment Uniques

      l += j; // Accumumlate cross check total

      Bins[j]++; // Increment bining for the last group of matches

#ifdef DEBUG
      if (j > 1) // Display when more than 1 instance
        printf("%3d %016I64X %s\n", j, x, DecodeVec(x));
#endif

      x = *BitVector; // Current value

      j = 1; // Reset match count to the single entry we have
    }

    BitVector++;
  }

#ifdef DEBUG
  if (j > 1) // Display when more than 1 instance
    printf("%3d %016I64X %s\n", j, x, DecodeVec(x));
#endif

  k++; // Account for pending uniques

  l += j; // Accumulate cross check total

  Bins[j]++; // Increment bining for the last group

#ifdef DEBUG
  putchar('\n');
#endif

  TSCEnd = ReadTSC(); // Processor Cycles
  TickEnd = GetTickCount(); // Clock Ticks (milliseconds)

  printf("%10d Ticks, %10d Cycles\n",TickEnd - TickStart, TSCEnd - TSCStart);

  putchar('\n');

  printf("%8d Unique\n",k);
  printf("%8d Elements\n",l); // Cross check, should match ELEMENTS

  putchar('\n');

  // Dump out Match count bining

  j = 0; // Cross of total (crosschecking)

  for(i=0; i<MATCHBINS; i++)
  {
    if (Bins[i]) // Output count of matches where bin contains a count
    {
      printf("Groups occuring %d times : %8d\n",i,Bins[i]);

      j += Bins[i]; // Accumulate cross check total
    }
  }

  printf("                          %8d\n",j); // Cross check, should match Uniques

  putchar('\n');

  // Generate some total timings

  printf("Entire Application Totals\n\n");

  TotalTSCEnd = ReadTSC(); // Processor Cycles - Not useful over long period
  TotalTickEnd = GetTickCount(); // Clock Ticks (milliseconds)

  printf("%10d Ticks, %10d Cycles\n",TotalTickEnd - TotalTickStart, TotalTSCEnd - TotalTSCStart);

  if ((TotalTickEnd - TotalTickStart) > 1000)
    puts("WARNING: Cycles overflowed or not useful");

  return(1);
}

//*****************************************************************************

